use std::str::Chars;

/// The different kinds of tokens for later consumption by the parser.
#[derive(Debug)]
pub enum TokenKind {
    // Command mode
    /// Any non-newline whitespace.
    Whitespace,
    /// One or more newline characters.
    Newlines,
    /// A semicolon.
    Semicolon,
    /// A pipe.
    Pipe,
    /// Double quotes.
    DoubleQuote,
    /// Plain text. Either outside or inside of quotes.
    PlainText,
    /// A backslash followed by another character.
    EscapedChar,
    /// A line comment, from # to newline.
    Comment,

    /// End of file.
    Eof,
}

/// Tokens represent the lowest abstraction over the input string.
/// They only contain their length, not their position, or text.
/// The parser is responsible for getting that information, if required.
#[derive(Debug)]
pub struct Token {
    pub kind: TokenKind,
    pub len: usize,
}

/// The lexer splits up a &str into a stream of tokens.
/// The lexer for oyster specifically supports multiple entry points,
/// which each parse a different sublanguage.
/// This makes handling strings, and substitutions in strings, and strings in substitutions in
/// strings way easier.
pub struct Lexer<'a> {
    chars: Chars<'a>,
    remaining: usize,
}

impl Lexer<'_> {
    /// Create a new lexer.
    pub fn new(source: &str) -> Lexer {
        let chars = source.chars();
        let remaining = chars.as_str().len();

        Lexer { chars, remaining }
    }

    /// Peek at the next character in the input stream.
    fn peek_char(&mut self) -> Option<char> {
        self.chars.clone().next()
    }

    /// Retrieve the next character from the input stream.
    fn next_char(&mut self) -> Option<char> {
        self.chars.next()
    }

    /// Consume input while `pred` holds true.
    fn eat_while(&mut self, pred: fn(char) -> bool) {
        loop {
            match self.peek_char() {
                Some(c) => {
                    if !pred(c) {
                        return;
                    }
                }
                None => return,
            }

            self.next_char();
        }
    }

    /// Returns the length of the current token.
    /// Also updates remaining characters in input.
    fn token_len(&mut self) -> usize {
        let old = self.remaining;
        self.remaining = self.chars.as_str().len();
        old - self.remaining
    }

    /// Get the next token in the command sublanguage.
    pub fn next_command_token(&mut self) -> Token {
        fn is_whitespace(c: char) -> bool {
            [' ', '\t'].contains(&c)
        }

        match self.next_char() {
            None => Token {
                kind: TokenKind::Eof,
                len: 0,
            },
            Some(c) => {
                let kind = match c {
                    ';' => TokenKind::Semicolon,
                    '|' => TokenKind::Pipe,
                    '"' => TokenKind::DoubleQuote,

                    c if is_whitespace(c) => {
                        self.eat_while(is_whitespace);
                        TokenKind::Whitespace
                    }

                    '\n' => {
                        self.eat_while(|c| c == '\n');
                        TokenKind::Newlines
                    }

                    '#' => {
                        self.eat_while(|c| c != '\n');
                        TokenKind::Comment
                    }

                    '\\' => {
                        self.next_char();
                        TokenKind::EscapedChar
                    }

                    _ => {
                        self.eat_while(|c| ![' ', '\t', '\n', ';', '|', '\\', '"'].contains(&c));
                        TokenKind::PlainText
                    }
                };

                let len = self.token_len();

                Token { kind, len }
            }
        }
    }

    /// Get the next token in the string sublanguage.
    pub fn next_string_token(&mut self) -> Token {
        match self.next_char() {
            None => Token {
                kind: TokenKind::Eof,
                len: 0,
            },
            Some(c) => {
                let kind = match c {
                    '"' => TokenKind::DoubleQuote,

                    '\\' => {
                        self.next_char();
                        TokenKind::EscapedChar
                    }

                    _ => {
                        self.eat_while(|c| !['"', '\\'].contains(&c));
                        TokenKind::PlainText
                    }
                };

                let len = self.token_len();

                Token { kind, len }
            }
        }
    }
}
